#include "license.readme"
#include "license_myspell.readme"

#include <cstdlib>
#include <cstring>
#include <cstdio>

#include "affixmgr.hxx"
#include "affentry.hxx"

#ifndef WINDOWS
using namespace std;
#endif


// First some base level utility routines
extern void   mychomp(char * s);                          // remove end of line char(s)
extern char * mystrdup(const char * s);                   // duplicate string
extern char * myrevstrdup(const char * s);                // duplicate reverse of string
extern char * mystrsep(char ** sptr, const char delim);   // parse into tokens with char delimiter
extern int    isSubset(const char * s1, const char * s2); // is affix s1 is a "subset" affix s2
extern int    isRevSubset(const char * s1, const char * s2, int len); // is affix s1 is a "subset" affix s2

AffixMgr::AffixMgr(const char * affpath, HashMgr* ptr) 
{
  // register hash manager and load affix data from aff file
  pHMgr = ptr;
  trystring = NULL;
  encoding=NULL;
  reptable = NULL;
  numrep = 0;
  compound = FLAG_NULL; // may word in compound forms
  compound2 = FLAG_NULL; // may first word in compound forms (HU)
  compound3 = FLAG_NULL; // may last word in compound forms (HU)
  compoundword = FLAG_NULL; // compound word signing flag (HU)
  compoundforbidflag = FLAG_NULL; // compound fordidden flag for suffixed word (HU)
  forbiddenword = FLAG_NULL; // forbidden word signing flag (HU)
  hu_kotohangzo = FLAG_NULL; // plus (``joining'') vowel sign (HU)
  onlyroot = FLAG_NULL; // forbidden word modify flag (HU)
  cpdwordmax=0; // default: unlimited wordcount in compound words (HU)
  cpdmin = 3;  // default value
  cpdmaxsyllable = 0; // default: unlimited syllablecount in compound words (HU)
  cpdvowels=NULL; // vowels (HU) (possible optimalization: charvector)
  pfxappnd=NULL; // previous prefix for counting the syllables of prefix (HU)
  sfxappnd=NULL; // previous suffix for counting a special syllables (HU)
  cpdsyllablenum=NULL; // syllable count incrementing flag (HU)
  accent=NULL; // accented letters (HU)
  checknum=0; // checking numbers, and word with numbers (HU)
  wordchars=NULL; // letters + spec. word characters (HU)
  version=NULL; // affix and dictionary file version string (HU)
  havecontclass=0; // flags of possible continuing classes (double affix)
  
  for (int i=0; i < SETSIZE; i++) {
     pStart[i] = NULL;
     sStart[i] = NULL;
     contclasses[i] = 0;
  }
  if (parse_file(affpath)) {
     fprintf(stderr,"Failure loading aff file %s\n",affpath);
     fflush(stderr);
  }
}


AffixMgr::~AffixMgr() 
{
 
  // pass through linked prefix entries and clean up
  for (int i=0; i < SETSIZE ;i++) {
       PfxEntry * ptr = (PfxEntry *)pStart[i];
       PfxEntry * nptr = NULL;
       while (ptr) {
            nptr = ptr->getNext();
            delete(ptr);
            ptr = nptr;
            nptr = NULL;
       }  
  }

  // pass through linked suffix entries and clean up
  for (int j=0; j < SETSIZE ; j++) {
       SfxEntry * ptr = (SfxEntry *)sStart[j];
       SfxEntry * nptr = NULL;
       while (ptr) {
            nptr = ptr->getNext();
            delete(ptr);
            ptr = nptr;
            nptr = NULL;
       }
       sStart[j] = NULL;
  }

  if (trystring) free(trystring);
  trystring=NULL;
  if (encoding) free(encoding);
  encoding=NULL;
  if (reptable) {  
     for (int j=0; j < numrep; j++) {
        free(reptable[j].pattern);
        free(reptable[j].replacement);
        reptable[j].pattern = NULL;
        reptable[j].replacement = NULL;
     }
     free(reptable);  
     reptable = NULL;
  }
  numrep = 0;
  FREE_FLAG(compound);
  FREE_FLAG(compound2);
  FREE_FLAG(compound3);
  FREE_FLAG(compoundforbidflag);
  FREE_FLAG(compoundword);
  FREE_FLAG(forbiddenword);
  FREE_FLAG(hu_kotohangzo);
  FREE_FLAG(onlyroot);
  cpdwordmax = 0;
  pHMgr = NULL;
  cpdmin = 0;
  cpdmaxsyllable = 0;
  if (cpdvowels) free(cpdvowels);
  if (cpdsyllablenum) free(cpdsyllablenum);
  if (accent) {
      free(accent->pattern);
      free(accent->replacement);
      accent->pattern = NULL;
      accent->replacement = NULL;
      free(accent);
  }
  if (wordchars) free(wordchars);
  if (version) free(version);
  checknum=0;

}


// read in aff file and build up prefix and suffix entry objects 
int  AffixMgr::parse_file(const char * affpath)
{

  // io buffers
  char line[MAXLNLEN+1];
 
  // affix type
  char ft;

  // open the affix file
  FILE * afflst;
  afflst = fopen(affpath,"r");
  if (!afflst) {
    fprintf(stderr,"Error - could not open affix description file %s\n",affpath);
    return 1;
  }

  // step one is to parse the affix file building up the internal
  // affix data structures


    // read in each line ignoring any that do not
    // start with a known line type indicator

    while (fgets(line,MAXLNLEN,afflst)) {
       mychomp(line);

       /* parse in the try string */
       if (strncmp(line,"TRY",3) == 0) {
          if (parse_try(line)) {
             return 1;
          }
       }

       /* parse in the name of the character set used by the .dict and .aff */
       if (strncmp(line,"SET",3) == 0) {
          if (parse_set(line)) {
             return 1;
          }
       }

       /* parse in the flag used by the controlled compound words */
       if (strncmp(line,"COMPOUNDFLAG",12) == 0) {
          if (parse_cpdflag(line)) {
             return 1;
          }
       }

       /* parse in the flag used by the controlled compound words */
       if (strncmp(line,"COMPOUNDFIRST",13) == 0) {
          if (parse_cpdfirst(line)) {
             return 1;
          }
       }

       /* parse in the flag used by the controlled compound words */
       if (strncmp(line,"COMPOUNDLAST",12) == 0) {
          if (parse_cpdlast(line)) {
             return 1;
          }
       }

       /* parse in the flag used by compound_check() method */
       if (strncmp(line,"COMPOUNDWORD",12) == 0) {
          if (parse_cpdword(line)) {
             return 1;
          }
       }

       /* parse in the flag used by compound_check() method */
       if (strncmp(line,"COMPOUNDFORBIDFLAG",18) == 0) {
          if (parse_cpdforbid(line)) {
             return 1;
          }
       }

       /* parse in the flag used by forbidden words */
       if (strncmp(line,"FORBIDDENWORD",13) == 0) {
          if (parse_forbid(line)) {
             return 1;
          }
       }

       /* parse in the FORBIDDENWORD modify flag */
       if (strncmp(line,"ONLYROOT",8) == 0) {
          if (parse_onlyroot(line)) {
             return 1;
          }
       }

       /* parse in the flag used by the controlled compound words */
       if (strncmp(line,"COMPOUNDMIN",11) == 0) {
          if (parse_cpdmin(line)) {
             return 1;
          }
       }

       /* parse in the flag used by compound_check() method */
       if (strncmp(line,"COMPOUNDSYLLABLE",16) == 0) {
          if (parse_cpdsyllable(line)) {
             return 1;
          }
       }

       /* parse in the accent strings */
       if (strncmp(line,"ACCENT",6) == 0) {
          if (parse_accent(line)) {
             return 1;
          }
       }

       /* parse in the flag used by compound_check() method */
       if (strncmp(line,"SYLLABLENUM",11) == 0) {
          if (parse_syllablenum(line)) {
             return 1;
          }
       }

       /* parse in the flag used by the controlled compound words */
       if (strncmp(line,"CHECKNUM",8) == 0) {
           checknum=1;
       }

       /* parse in the try string */
       if (strncmp(line,"WORDCHARS",9) == 0) {
          if (parse_wordchars(line)) {
             return 1;
          }
       }

       /* parse in the typical fault correcting table */
       if (strncmp(line,"REP",3) == 0) {
          if (parse_reptable(line, afflst)) {
             return 1;
          }
       }

       /* parse in the flag used by forbidden words */
       if (strncmp(line,"HU_KOTOHANGZO",13) == 0) {
          if (parse_hu_kotohangzo(line)) {
             return 1;
          }
       }

       /* parse in the version string */
       if (strncmp(line,"VERSION",7) == 0) {
          if (parse_version(line)) {
             return 1;
          }
       }


       // parse this affix: P - prefix, S - suffix
       ft = ' ';
       if (strncmp(line,"PFX",3) == 0) ft = 'P';
       if (strncmp(line,"SFX",3) == 0) ft = 'S';
       if (ft != ' ') {
          if (parse_affix(line, ft, afflst)) {
             return 1;
          }
       }

    }
    fclose(afflst);
    
    // quick loading tree convert to list

    process_sfx_tree_to_list();

    // now we can speed up performance greatly taking advantage of the 
    // relationship between the affixes and the idea of "subsets".

    // View each prefix as a potential leading subset of another and view
    // each suffix (reversed) as a potential trailing subset of another.

    // To illustrate this relationship if we know the prefix "ab" is found in the
    // word to examine, only prefixes that "ab" is a leading subset of need be examined.
    // Furthermore is "ab" is not present then none of the prefixes that "ab" is
    // is a subset need be examined.
    // The same argument goes for suffix string that are reversed.

    // Then to top this off why not examine the first char of the word to quickly
    // limit the set of prefixes to examine (i.e. the prefixes to examine must 
    // be leading supersets of the first character of the word (if they exist)
 
    // To take advantage of this "subset" relationship, we need to add two links
    // from entry.  One to take next if the current prefix is found (call it nexteq)
    // and one to take next if the current prefix is not found (call it nextne).

    // Since we have built ordered lists, all that remains is to properly intialize 
    // the nextne and nexteq pointers that relate them

    process_pfx_order();
    process_sfx_order();

    // expand wordchars string (HU), based on csutil

    struct cs_info * csconv;
    char * enc = get_encoding();
    csconv = get_current_cs(enc);
    free(enc);
    enc = NULL;

    char expw[MAXLNLEN];
    if (wordchars) {
	strcpy(expw, wordchars);
	free(wordchars);
    } else *expw = '\0';

    for (int i = 0; i <= 255; i++) {
	if ( (csconv[i].cupper != csconv[i].clower) &&
	    (! strchr(expw, (char) i))) {
		*(expw + strlen(expw) + 1) = '\0';
		*(expw + strlen(expw)) = (char) i;
	    }
    }

    wordchars = mystrdup(expw);

    return 0;
}


int AffixMgr::build_pfxlist(AffEntry* pfxptr)
{
  PfxEntry * ptr;
  PfxEntry * pptr;
  PfxEntry * ep = (PfxEntry*) pfxptr;

  // get the right starting points
  const char * key = ep->getKey();

  // next index by affix string

  // handle the special case of null affix string
  if (strlen(key) == 0) {
    // always inset them at head of list at element 0
     ptr = (PfxEntry*)pStart[0];
     ep->setNext(ptr);
     pStart[0] = (AffEntry*)ep;
     return 0;
  }

  // now handle the general case
  unsigned char sp = *((const unsigned char *)key);
  ptr = (PfxEntry*)pStart[sp];

  /* handle the insert at top of list case */
  if ((!ptr) || ( strcmp( ep->getKey() , ptr->getKey() ) <= 0)) {
     ep->setNext(ptr);
     pStart[sp] = (AffEntry*)ep;
     return 0;
  }

  /* otherwise find where it fits in order and insert it */
  pptr = NULL;
  for (; ptr != NULL; ptr = ptr->getNext()) {
    if (strcmp( ep->getKey() , ptr->getKey() ) <= 0) break;
    pptr = ptr;
  }
  pptr->setNext(ep);
  ep->setNext(ptr);
  return 0;
}

// build an ordered tree of suffix entries (affix reversed)
int AffixMgr::build_sfxtree(AffEntry* sfxptr)
{
  SfxEntry * ptr;
  SfxEntry * pptr;
  SfxEntry * ep = (SfxEntry *) sfxptr;

  /* get the right starting point */
  const char * key = ep->getKey();

  // handle the special case of null affix string
  if (strlen(key) == 0) {
    // always inset them at head of list at element 0
     ptr = (SfxEntry*)sStart[0];
     ep->setNext(ptr);
     sStart[0] = (AffEntry*)ep;
     return 0;
  }

  // now handle the normal case
  unsigned char sp = *((const unsigned char *)key);
  ptr = (SfxEntry*)sStart[sp];
  
  /* handle the insert at top of tree case */

  ep->setNextEQ(NULL);
  ep->setNextNE(NULL);

  if (!ptr) {
     sStart[sp] = (AffEntry*)ep;
     return 0;
  }

  /* otherwise find where it fits in order and insert it into tree */

  pptr = NULL;
  for (;;) {
    pptr = ptr;
    if (strcmp( ep->getKey(), ptr->getKey() ) <= 0) {
        ptr = ptr->getNextEQ();
        if (!ptr) {
            pptr->setNextEQ(ep);
            break;
        }
    } else {
        ptr = ptr->getNextNE();    
        if (!ptr) {
            pptr->setNextNE(ep);
            break;
        }
    }
  }
  return 0;
}



// initialize the PfxEntry links NextEQ and NextNE to speed searching
int AffixMgr::process_pfx_order()
{
    PfxEntry* ptr;

    // loop through each prefix list starting point
    for (int i=1; i < SETSIZE; i++) {

         ptr = (PfxEntry*)pStart[i];

         // look through the remainder of the list
         //  and find next entry with affix that 
         // the current one is not a subset of
         // mark that as destination for NextNE
         // use next in list that you are a subset
         // of as NextEQ

         for (; ptr != NULL; ptr = ptr->getNext()) {

	     PfxEntry * nptr = ptr->getNext();
             for (; nptr != NULL; nptr = nptr->getNext()) {
	         if (! isSubset( ptr->getKey() , nptr->getKey() )) break;
             }
             ptr->setNextNE(nptr);
             ptr->setNextEQ(NULL);
             if ((ptr->getNext()) && isSubset(ptr->getKey() , (ptr->getNext())->getKey())) 
                 ptr->setNextEQ(ptr->getNext());
         }

         // now clean up by adding smart search termination strings:
         // if you are already a superset of the previous prefix
         // but not a subset of the next, search can end here
         // so set NextNE properly

         ptr = (PfxEntry *) pStart[i];
         for (; ptr != NULL; ptr = ptr->getNext()) {
	     PfxEntry * nptr = ptr->getNext();
             PfxEntry * mptr = NULL;
             for (; nptr != NULL; nptr = nptr->getNext()) {
	         if (! isSubset(ptr->getKey(),nptr->getKey())) break;
                 mptr = nptr;
             }
             if (mptr) mptr->setNextNE(NULL);
         }
    }
    return 0;
}

// convert SfxEntry tree to list with inorder tree walking
AffEntry * AffixMgr::process_sfx_inorder(AffEntry* ptr, AffEntry* nptr) {
    if (ptr) {
        nptr = process_sfx_inorder(((SfxEntry*) ptr)->getNextNE(), nptr);
        ((SfxEntry*) ptr)->setNext((SfxEntry*) nptr);
        nptr = process_sfx_inorder(((SfxEntry*) ptr)->getNextEQ(), ptr);
    } 
    return nptr;
}

// convert SfxEntry trees to lists with inorder tree walking
int AffixMgr::process_sfx_tree_to_list()
{
    for (int i=1; i < SETSIZE; i++) {
         sStart[i] = process_sfx_inorder(sStart[i], NULL);
    }    
    return 0;
}

// initialize the SfxEntry links NextEQ and NextNE to speed searching
int AffixMgr::process_sfx_order()
{
    SfxEntry* ptr;

    // loop through each prefix list starting point
    for (int i=1; i < SETSIZE; i++) {

         ptr = (SfxEntry *) sStart[i];

         // look through the remainder of the list
         //  and find next entry with affix that 
         // the current one is not a subset of
         // mark that as destination for NextNE
         // use next in list that you are a subset
         // of as NextEQ

         for (; ptr != NULL; ptr = ptr->getNext()) {
	     SfxEntry * nptr = ptr->getNext();
             for (; nptr != NULL; nptr = nptr->getNext()) {
	         if (! isSubset(ptr->getKey(),nptr->getKey())) break;
             }
             ptr->setNextNE(nptr);
             ptr->setNextEQ(NULL);
             if ((ptr->getNext()) && isSubset(ptr->getKey(),(ptr->getNext())->getKey())) 
                 ptr->setNextEQ(ptr->getNext());
         }


         // now clean up by adding smart search termination strings:
         // if you are already a superset of the previous suffix
         // but not a subset of the next, search can end here
         // so set NextNE properly

         ptr = (SfxEntry *) sStart[i];
         for (; ptr != NULL; ptr = ptr->getNext()) {
	     SfxEntry * nptr = ptr->getNext();
             SfxEntry * mptr = NULL;
             for (; nptr != NULL; nptr = nptr->getNext()) {
	         if (! isSubset(ptr->getKey(),nptr->getKey())) break;
                 mptr = nptr;
             }
             if (mptr) mptr->setNextNE(NULL);
         }
    }
    return 0;
}



// takes aff file condition string and creates the
// conds array - please see the appendix at the end of the
// file affentry.cxx which describes what is going on here
// in much more detail

void AffixMgr::encodeit(struct affentry * ptr, char * cs)
{
  unsigned char c;
  int i, j, k;
  unsigned char mbr[MAXLNLEN];

  // now clear the conditions array */
  for (i=0;i<SETSIZE;i++) ptr->conds[i] = (unsigned char) 0;

  // now parse the string to create the conds array */
  int nc = strlen(cs);
  int neg = 0;   // complement indicator
  int grp = 0;   // group indicator
  int n = 0;     // number of conditions
  int ec = 0;    // end condition indicator
  int nm = 0;    // number of member in group

  // if no condition just return
  if (strcmp(cs,".")==0) {
    ptr->numconds = 0;
    return;
  }

  i = 0;
  while (i < nc) {
    c = *((unsigned char *)(cs + i));

    // start group indicator
    if (c == '[') {
       grp = 1;
       c = 0;
    }

    // complement flag
    if ((grp == 1) && (c == '^')) {
       neg = 1;
       c = 0;
    }

    // end goup indicator
    if (c == ']') {
       ec = 1;
       c = 0;
    }

    // add character of group to list
    if ((grp == 1) && (c != 0)) {
      *(mbr + nm) = c;
      nm++;
      c = 0;
    }

    // end of condition 
    if (c != 0) {
       ec = 1;
    }

    
    if (ec) {
      if (grp == 1) {
        if (neg == 0) {
          // set the proper bits in the condition array vals for those chars
	  for (j=0;j<nm;j++) {
	     k = (unsigned int) mbr[j];
             ptr->conds[k] = ptr->conds[k] | (1 << n);
          }
	} else {
	  // complement so set all of them and then unset indicated ones
	   for (j=0;j<SETSIZE;j++) ptr->conds[j] = ptr->conds[j] | (1 << n);
	   for (j=0;j<nm;j++) {
	     k = (unsigned int) mbr[j];
             ptr->conds[k] = ptr->conds[k] & ~(1 << n);
	   }
        }
        neg = 0;
        grp = 0;   
        nm = 0;
      } else {
         // not a group so just set the proper bit for this char
         // but first handle special case of . inside condition
         if (c == '.') {
	    // wild card character so set them all
            for (j=0;j<SETSIZE;j++) ptr->conds[j] = ptr->conds[j] | (1 << n);
         } else {  
	    ptr->conds[(unsigned int) c] = ptr->conds[(unsigned int)c] | (1 << n);
         }
      }
      n++;
      ec = 0;
    }


    i++;
  }
  ptr->numconds = n;
  return;
}


// check word for prefixes
struct hentry * AffixMgr::prefix_check(const char * word, int len, int prefix_in_cpms,
    const FLAG needflag)
{
    struct hentry * rv= NULL;

    pfxappnd = NULL;
    sfxappnd = NULL;
    
    // first handle the special case of 0 length prefixes
    PfxEntry * pe = (PfxEntry *) pStart[0];
    while (pe) {
       rv = pe->check(word, len, prefix_in_cpms, needflag);
       if (rv) return rv;
       pe = pe->getNext();
    }
  
    // now handle the general case
    unsigned char sp = *((const unsigned char *)word);
    PfxEntry * pptr = (PfxEntry *)pStart[sp];

    while (pptr) {
        if (isSubset(pptr->getKey(),word)) {
	     rv = pptr->check(word, len, prefix_in_cpms, needflag);
             if (rv) {
                pfx=(AffEntry *)pptr; // BUG: sfx not stateless
	        return rv;
	     }
             pptr = pptr->getNextEQ();
        } else {
	     pptr = pptr->getNextNE();
        }
    }
    
    return NULL;
}

// check word for prefixes
struct hentry * AffixMgr::prefix_check_twosfx(const char * word, int len,
    int prefix_in_cpms, const FLAG needflag)
{
    struct hentry * rv= NULL;

    pfx = NULL;
    sfxappnd = NULL;
    
    // first handle the special case of 0 length prefixes
    PfxEntry * pe = (PfxEntry *) pStart[0];
    
    while (pe) {
#ifndef EXT_CLASS
        if ((contclasses[(unsigned int)(pe->getFlag())]))
#endif
        {
            rv = pe->check_twosfx(word, len, prefix_in_cpms, needflag);
            if (rv) return rv;
        }
        pe = pe->getNext();
    }
  
    // now handle the general case
    unsigned char sp = *((const unsigned char *)word);
    PfxEntry * pptr = (PfxEntry *)pStart[sp];

    while (pptr) {
        if (isSubset(pptr->getKey(),word)) {
#ifndef EXT_CLASS
            if (contclasses[(unsigned int)(pptr->getFlag())])
#endif
            {
                rv = pptr->check_twosfx(word, len, prefix_in_cpms, needflag);
                if (rv) {
                    pfx = (AffEntry *)pptr;
	            return rv;
	        }
            }
            pptr = pptr->getNextEQ();
        } else {
	     pptr = pptr->getNextNE();
        }
    }
    
    return NULL;
}

#ifdef HUNMORPH
// check word for prefixes
char * AffixMgr::prefix_check_morph(const char * word, int len, int prefix_in_cpms,
    const FLAG needflag)
{
    char * st;

    char result[MAXLNLEN];
    result[0] = '\0';

    pfx = NULL;
    sfxappnd = NULL;
    
    // first handle the special case of 0 length prefixes
    PfxEntry * pe = (PfxEntry *) pStart[0];
    while (pe) {
       st = pe->check_morph(word,len,prefix_in_cpms, needflag);
       if (st) {
            strcat(result, st);
            free(st);
       }
       // if (rv) return rv;
       pe = pe->getNext();
    }
  
    // now handle the general case
    unsigned char sp = *((const unsigned char *)word);
    PfxEntry * pptr = (PfxEntry *)pStart[sp];

    while (pptr) {
        if (isSubset(pptr->getKey(),word)) {
            st = pptr->check_morph(word,len,prefix_in_cpms, needflag);
            if (st) {
                strcat(result, st);
                free(st);
                pfx = (AffEntry *)pptr;
            }
            pptr = pptr->getNextEQ();
        } else {
	    pptr = pptr->getNextNE();
        }
    }
    
    if (*result) return mystrdup(result);
    return NULL;
}


// check word for prefixes
char * AffixMgr::prefix_check_twosfx_morph(const char * word, int len,
    int prefix_in_cpms, const FLAG needflag)
{
    char * st;

    char result[MAXLNLEN];
    result[0] = '\0';

    pfx = NULL;
    sfxappnd = NULL;
    
    // first handle the special case of 0 length prefixes
    PfxEntry * pe = (PfxEntry *) pStart[0];
    while (pe) {
#ifndef EXT_CLASS
        if ((contclasses[(unsigned int)(pe->getFlag())]))
#endif
        {
            st = pe->check_twosfx_morph(word,len,prefix_in_cpms, needflag);
            if (st) {
                strcat(result, st);
                free(st);
            }
        }
       pe = pe->getNext();
    }
  
    // now handle the general case
    unsigned char sp = *((const unsigned char *)word);
    PfxEntry * pptr = (PfxEntry *)pStart[sp];

    while (pptr) {
        if (isSubset(pptr->getKey(),word)) {
#ifndef EXT_CLASS
            if (contclasses[(unsigned int)(pptr->getFlag())])
#endif
            {
                st = pptr->check_twosfx_morph(word, len, prefix_in_cpms, needflag);
                if (st) {
                    strcat(result, st);
                    free(st);
                    pfx = (AffEntry *)pptr;
                }
            }
            pptr = pptr->getNextEQ();
        } else {
	    pptr = pptr->getNextNE();
        }
    }
    
    if (*result) return mystrdup(result);
    return NULL;
}
#endif

// check compounds with replacement table
// rationale: a typical fault may be a `right' compound word,
// with this function we correct a sort of typical hungarian faults:
// for example:
// service = szerviz
// drug+water = szervz, this form was `right' compound.
int AffixMgr::repl_check(const char * word, int wl)
{
  char candidate[MAXLNLEN];
  const char * r;
  int lenr, lenp;

  if (wl < 2) return 0;

  int numrep = get_numrep();
  struct replentry* reptable = get_reptable();
  if (reptable==NULL) return 0;

  for (int i=0; i < numrep; i++ ) {
      r = word;
      lenr = strlen(reptable[i].replacement);
      lenp = strlen(reptable[i].pattern);
      // search every occurence of the pattern in the word
      while ((r=strstr(r, reptable[i].pattern)) != NULL) {
	  strcpy(candidate, word);
	  if (r-word + lenr + strlen(r+lenp) >= MAXLNLEN) break;
	  strcpy(candidate+(r-word),reptable[i].replacement);
	  strcpy(candidate+(r-word)+lenr, r+lenp);
          if (candidate_check(candidate,strlen(candidate))) return 1;
          if (candidate_check(candidate,strlen(candidate))) return 1;
          r++; // search for the next letter
      }
   }
   return 0;
}

#if 0
// check compounds with adjacent letter were swapped
int AffixMgr::swap_check(const char * word, int wl)
{
   char	candidate[MAXLNLEN];
   char * p;
   char	tmpc;

   if (wl < 2) return 0;

   // try swapping adjacent chars one by one
   strcpy(candidate, word);
   for (p = candidate;  p[1] != 0;  p++) {
      tmpc = *p;
      *p = p[1];
      p[1] = tmpc;
      if (candidate_check(candidate,wl)) return 1;
      tmpc = *p;
      *p = p[1];
      p[1] = tmpc;
   }
   return 0;
}
#endif

inline int AffixMgr::candidate_check(const char * word, int len)
{
  struct hentry * rv=NULL;
  
  rv = lookup(word);
  if (rv) return 1;

//  rv = prefix_check(word,len,1);
//  if (rv) return 1;
  
  rv = affix_check(word,len);
  if (rv) return 1;
  return 0;
}

// calculate number of syllable for compound-checking
int AffixMgr::get_syllable(const char * word, int wlen)
{
    if (cpdmaxsyllable==0) return 0;
    
    int num=0;

    for (int i=0; i<wlen; i++) {
	if (strchr(cpdvowels, word[i])) num++;
    }
    return num;	
}

// check if compound word is correctly spelled
// hu_mov_rule = mozgszably
struct hentry * AffixMgr::compound_check(const char * word, int len, 
    int wordnum, int numsyllable, int maxwordnum,
    int hu_mov_rule = 0, int * cmpdstemnum = NULL, int * cmpdstem = NULL)
{
    int i, oldnumsyllable, oldnumsyllable2, oldwordnum, oldwordnum2;
    int oldcmpdstemnum = 0;
    struct hentry * rv = NULL;
    struct hentry * rv_first;
    char st [MAXWORDLEN + 1];
    char ch;
    
    int checked_prefix;

#ifdef HUNSTEM
    if (cmpdstemnum) {
	if (wordnum == 0) {
	    *cmpdstemnum = 1;
	} else {
	    (*cmpdstemnum)++;
	}
    }
#endif

    // handle case of string too short to be a piece of a compound word 
    if (len < cpdmin) return NULL;

    strcpy(st, word);
    
    for (i=cpdmin; i < (len - (cpdmin-1)); i++) {
	oldnumsyllable = numsyllable;
	oldwordnum = wordnum;
        checked_prefix = 0;
	
        ch = st[i];
	st[i] = '\0';
        sfx = NULL;

	// FIRST WORD
	
	rv = lookup(st); // perhaps without prefix

        if (!rv) {
            rv = prefix_check(st, i, 0, compound);    // or with prefix
            if (!rv) {
                if ((rv = suffix_check(st, i, 0, NULL, NULL, 0, NULL, FLAG_NULL, compound, 0)) &&
                    ((SfxEntry*)sfx)->getCont() &&
                        (TESTAFF(((SfxEntry*)sfx)->getCont(), compoundforbidflag, 
                            strlen(((SfxEntry*)sfx)->getCont())) ||
                        TESTAFF(((SfxEntry*)sfx)->getCont(), compound3, 
                            strlen(((SfxEntry*)sfx)->getCont())))) {
                        rv = NULL;
                }
            }
            //if (!rv) rv = suffix_check(st, i, 0, NULL, NULL, 0, NULL, '\0', compound2, 0);
            if (rv) checked_prefix = 1;
        // else check forbiddenwords
	} else if ((rv->astr) && (TESTAFF(rv->astr, forbiddenword, rv->alen))) {
		st[i] = ch;
		continue;
	}

	// check forbiddenwords
	if ((rv) && (rv->astr) && (TESTAFF(rv->astr, forbiddenword, rv->alen))
	    && (! TESTAFF(rv->astr, onlyroot, rv->alen))) return NULL;

	// increment word number, if the second root has a compoundword flag
	if ((rv) && (compoundword) && 
	    (TESTAFF(rv->astr, compoundword, rv->alen))) {
		wordnum++;
	}

	// first word is acceptable in compound words?
	if (((rv) && 
	  ( checked_prefix ||
	    TESTAFF(rv->astr, compound, rv->alen) ||
	    TESTAFF(rv->astr, compound2, rv->alen) 
#ifndef EXT_CLASS
	    || (	// hu_mov_rule
	        hu_mov_rule && (
		    TESTAFF(rv->astr, 'F', rv->alen) ||
		    TESTAFF(rv->astr, 'G', rv->alen) ||
		    TESTAFF(rv->astr, 'H', rv->alen)
		)
	      )
#endif
	  )
	  && ! (( // test triple letters
                   (word[i-1]==word[i]) && (
                      ((i>1) && (word[i-1]==word[i-2])) || 
                      ((word[i-1]==word[i+1])) // may word[i+1] == '\0'
		   )
               ) ||
	       ( // test hungarian consonant
	           (i>2) && (
		       (strncmp(word+i-3,"ccscs",5)==0) || // giccscsinls
		       (strncmp(word+i-3,"ggygy",5)==0) || // meggygyjts
		       (strncmp(word+i-3,"llyly",5)==0) || // gallylyukaszts
		       (strncmp(word+i-3,"nnyny",5)==0) || // knnynyals
	               (strncmp(word+i-3,"sszsz",5)==0) || // dzsesszszak (hibs plda)
		       (strncmp(word+i-2,"szszer",7)==0)
		       //(strncmp(word+i-3,"ttyty",5)==0) // may word[i+1] == '\0'
		   )
	       ))       
         )
#ifndef EXT_CLASS
         || ((!rv) && hu_mov_rule && (rv = affix_check(st,i)) &&
              (sfx && ((SfxEntry*)sfx)->getCont() && (
                        TESTAFF(((SfxEntry*)sfx)->getCont(), 'x', strlen(((SfxEntry*)sfx)->getCont())) ||
                        TESTAFF(((SfxEntry*)sfx)->getCont(), '%', strlen(((SfxEntry*)sfx)->getCont()))
                    )                
               )
	     )
#endif
         ) {

	    // calculate syllable number of the word
	    numsyllable += get_syllable(st, i);

	    // + 1 word, if syllable number of the prefix > 1 (hungarian convention)
	    if (pfx && (get_syllable(((PfxEntry *)pfx)->getKey(),strlen(((PfxEntry *)pfx)->getKey())) > 1)) wordnum++;

#ifdef HUNSTEM
	    if (cmpdstem) cmpdstem[*cmpdstemnum - 1] = i;
#endif

	    // NEXT WORD(S)
	    rv_first = rv;
	    rv = lookup((word+i)); // perhaps without prefix

        // search homonym with compound flag
        while ((rv) && (rv->next_homonym) && 
			!((TESTAFF(rv->astr, compound, rv->alen)) ||
			  (compound3 && (TESTAFF(rv->astr, compound3, rv->alen))))) {
            rv = rv->next_homonym;
        }

	    oldnumsyllable2 = numsyllable;
	    oldwordnum2 = wordnum;

	    // fonev_morfo
#ifndef EXT_CLASS
	    if ((rv) && (TESTAFF(rv->astr, 'I', rv->alen)) && !(TESTAFF(rv->astr, 'J', rv->alen))) {
		numsyllable--;
	    }
#endif
	    // increment word number, if the second root has a compoundword flag
	    if ((rv) && (compoundword) && 
		(TESTAFF(rv->astr, compoundword, rv->alen))) {
		    wordnum++;
	    }

	    // check forbiddenwords
	    if ((rv) && (rv->astr) && (TESTAFF(rv->astr,forbiddenword,rv->alen))
		    && (! TESTAFF(rv->astr, onlyroot, rv->alen))) return NULL;
	    // second word is acceptable, as a root?
	    // hungarian conventions: compounding is acceptable,
	    // when compound forms consist of 2 words, or if more,
	    // then the syllable number of root words must be 6, or lesser.

	    if ((rv) && (
	              TESTAFF(rv->astr, compound, rv->alen) ||
	              ((compound3 !=  FLAG_NULL) &&
		             TESTAFF(rv->astr, compound3, rv->alen))
	            )
		&& (
		      ((cpdwordmax==0) || (wordnum+1<cpdwordmax)) || 
		      ((cpdmaxsyllable==0) || 
		          (numsyllable+get_syllable(rv->word,rv->wlen)<=cpdmaxsyllable))
		    )
		&& (
		     (rv != rv_first)
		   )
		)
		 {
		      // bad compound word
		      if (repl_check(word,len)) return NULL;
		      return rv;
	    }

	    numsyllable = oldnumsyllable2 ;
	    wordnum = oldwordnum2;

	    // perhaps second word has prefix or/and suffix
            sfx = NULL;
	    sfxflag = FLAG_NULL;
	    rv = affix_check((word+i),strlen(word+i), compound);

            if (!rv && compound3) {
                sfx = NULL;
                pfx = NULL;
                rv = affix_check((word+i),strlen(word+i), compound3);
            }

            // check non_compound flag in suffix and prefix
            if ((rv) && 
                ((pfx && ((PfxEntry*)pfx)->getCont() &&
                    TESTAFF(((PfxEntry*)pfx)->getCont(), compoundforbidflag, 
                        strlen(((PfxEntry*)pfx)->getCont()))) ||
                (sfx && ((SfxEntry*)sfx)->getCont() &&
                    TESTAFF(((SfxEntry*)sfx)->getCont(), compoundforbidflag, 
                        strlen(((SfxEntry*)sfx)->getCont()))))) {
                    rv = NULL;
            }

	    // check forbiddenwords
	    if ((rv) && (rv->astr) && (TESTAFF(rv->astr,forbiddenword,rv->alen))
	    	    && (! TESTAFF(rv->astr, onlyroot, rv->alen))) return NULL;

	    // pfxappnd = prefix of word+i, or NULL
	    // calculate syllable number of prefix.
	    // hungarian convention: when syllable number of prefix is more,
	    // than 1, the prefix+word counts as two words.
//	    if (pfxappnd) {
//		ln = get_syllable(pfxappnd,strlen(pfxappnd));
//		numsyllable += ln;
//		if (ln>1) wordnum++;
//	    }

	    // calculate syllable number of the word
	    numsyllable += get_syllable(word + i, strlen(word + i));

            // - affix syllable num.
            if (sfxappnd) {
	        numsyllable -= get_syllable(sfxappnd, strlen(sfxappnd));
            }

	    // + 1 word, if syllable number of the prefix > 1 (hungarian convention)
	    if (pfx && (get_syllable(((PfxEntry *)pfx)->getKey(),strlen(((PfxEntry *)pfx)->getKey())) > 1)) wordnum++;

	    // increment syllable num, if last word has a SYLLABLENUM flag
	    // and the suffix is beginning `s'
            
#ifndef EXT_CLASS
	    if (cpdsyllablenum) {
	      switch (sfxflag) {
		case 'c': { numsyllable+=2; break; }
		case 'J': { numsyllable += 1; break; }
                case 'I': { if (TESTAFF(rv->astr, 'J', rv->alen)) numsyllable += 1; break; }
	      }
	    }
#endif	    
	    // increment word number, if the second word has a compoundword flag
	    if ((rv) && (compoundword) && 
		(TESTAFF(rv->astr, compoundword, rv->alen))) {
		    wordnum++;
	    }

	    // second word is acceptable, as a word with prefix or/and suffix?
	    // hungarian conventions: compounding is acceptable,
	    // when compound forms consist 2 word, otherwise
	    // the syllable number of root words is 6, or lesser.
	    if ((rv) && 
                    (
		      ((cpdwordmax ==0 ) || (wordnum + 1 < cpdwordmax)) || 
		      ((cpdmaxsyllable == 0) || 
		          (numsyllable <= cpdmaxsyllable))
		    )
		&& (
		   (rv != rv_first)
		   )) {
		    // bad compound word
		    if (repl_check(word, len)) return NULL;
		    return rv;
	    }

	    numsyllable = oldnumsyllable2;
	    wordnum = oldwordnum2;
#ifdef HUNSTEM
	    if (cmpdstemnum) oldcmpdstemnum = *cmpdstemnum;
#endif
	    // perhaps second word is a compound word (recursive call)
	    if (wordnum < maxwordnum) {
		rv = compound_check((word+i),strlen(word+i), wordnum+1,
		     numsyllable, maxwordnum, 0, cmpdstemnum, cmpdstem);
	    } else {
		rv=NULL;
	    }
	    if (rv) {
		// bad compound word
		if (repl_check(word, len)) return NULL;
		return rv;
	    } else {
#ifdef HUNSTEM
	    if (cmpdstemnum) *cmpdstemnum = oldcmpdstemnum;
#endif
	    }
	}
        st[i] = ch;
	wordnum = oldwordnum;
	numsyllable = oldnumsyllable;
    }
    return NULL;
}    

#ifdef HUNMORPH
// check if compound word is correctly spelled
// hu_mov_rule = mozgszably
int AffixMgr::compound_check_morph(const char * word, int len, 
    int wordnum, int numsyllable, int maxwordnum,
    int hu_mov_rule = 0, char ** result = NULL, char * partresult = NULL)
{
    int i, oldnumsyllable, oldnumsyllable2, oldwordnum, oldwordnum2;
	int ok = 0;

    struct hentry * rv = NULL;
    struct hentry * rv_first;
    char st [MAXWORDLEN + 1];
    char ch;
    
    int checked_prefix;
    char presult[MAXLNLEN];

    // handle case of string too short to be a piece of a compound word 
    if (len < cpdmin) return 0;

    strcpy(st, word);
    
    for (i=cpdmin; i < (len - (cpdmin-1)); i++) {
	oldnumsyllable = numsyllable;
	oldwordnum = wordnum;
        checked_prefix = 0;
	
        ch = st[i];
	st[i] = '\0';
        sfx = NULL;

	// FIRST WORD

        *presult = '\0';
        if (partresult) strcat(presult, partresult);
	
	rv = lookup(st); // perhaps without prefix

        if (rv)	 {
            if (rv->description) {
				if (*(rv->description) == '[') strcat(presult, st);
                strcat(presult, rv->description);
            } else {
            	strcat(presult, st);
                strcat(presult, "[STEM]");
            }
            strcat(presult, "+");
        }
        
        if (!rv) {
            rv = prefix_check(st, i, 0, compound);    // or with prefix
            if (!rv) {
                if ((rv = suffix_check(st, i, 0, NULL, NULL, 0, NULL, FLAG_NULL, compound, 0)) &&
                    ((SfxEntry*)sfx)->getCont() &&
                        (TESTAFF(((SfxEntry*)sfx)->getCont(), compoundforbidflag, 
                            strlen(((SfxEntry*)sfx)->getCont())) ||
                        TESTAFF(((SfxEntry*)sfx)->getCont(), compound3, 
                            strlen(((SfxEntry*)sfx)->getCont())))) {
                        rv = NULL;
                }
            }
            //if (!rv) rv = suffix_check(st, i, 0, NULL, NULL, 0, NULL, '\0', compound2, 0);
            if (rv) {
                //char * p = prefix_check_morph(st, i, 0, compound);
                char * p = affix_check_morph(st, i, compound);
                if (*p == '\0') p = affix_check_morph(st, i, compound2);
                if (*p == '\0') {
                    strcat(presult, "[STEM]");
                } else {
                    line_uniq(p);
                    if (strchr(p, '\n')) {
                        strcat(presult, "(");
                        strcat(presult, line_join(p, '|'));
                        strcat(presult, ")");
                      } else {
                        strcat(presult, p);
                      }
                }
                if (presult[strlen(presult) - 1] == '\n') {
                    presult[strlen(presult) - 1] = '\0';
                }
                checked_prefix = 1;
                strcat(presult, "+");
            }
        // else check forbiddenwords
	} else if ((rv->astr) && (TESTAFF(rv->astr, forbiddenword, rv->alen))) {
		st[i] = ch;
		continue;
	}
        
	// check forbiddenwords
	if ((rv) && (rv->astr) && (TESTAFF(rv->astr, forbiddenword, rv->alen))
	    && (! TESTAFF(rv->astr, onlyroot, rv->alen))) continue;

	// increment word number, if the second root has a compoundword flag
	if ((rv) && (compoundword) && 
	    (TESTAFF(rv->astr, compoundword, rv->alen))) {
		wordnum++;
	}

	// first word is acceptable in compound words?
	if (((rv) && 
	  ( checked_prefix ||
	    TESTAFF(rv->astr, compound, rv->alen) ||
	    TESTAFF(rv->astr, compound2, rv->alen) 
#ifndef EXT_CLASS
	    || (	// hu_mov_rule
	        hu_mov_rule && (
		    TESTAFF(rv->astr, 'F', rv->alen) ||
		    TESTAFF(rv->astr, 'G', rv->alen) ||
		    TESTAFF(rv->astr, 'H', rv->alen)
		)
	      )
#endif
	  )
	  && ! (( // test triple letters
                   (word[i-1]==word[i]) && (
                      ((i>1) && (word[i-1]==word[i-2])) || 
                      ((word[i-1]==word[i+1])) // may word[i+1] == '\0'
		   )
               ) ||
	       ( // test hungarian consonant
	           (i>2) && (
		       (strncmp(word+i-3,"ccscs",5)==0) || // giccscsinls
		       (strncmp(word+i-3,"ggygy",5)==0) || // meggygyjts
		       (strncmp(word+i-3,"llyly",5)==0) || // gallylyukaszts
		       (strncmp(word+i-3,"nnyny",5)==0) || // knnynyals
	               (strncmp(word+i-3,"sszsz",5)==0) || // dzsesszszak (hibs plda)
		       (strncmp(word+i-2,"szszer",7)==0)
		       //(strncmp(word+i-3,"ttyty",5)==0) // may word[i+1] == '\0'
		   )
	       ))       
         )
#ifndef EXT_CLASS
         || ((!rv) && hu_mov_rule && (rv = affix_check(st,i)) &&
              (sfx && ((SfxEntry*)sfx)->getCont() && (
                        TESTAFF(((SfxEntry*)sfx)->getCont(), 'x', strlen(((SfxEntry*)sfx)->getCont())) ||
                        TESTAFF(((SfxEntry*)sfx)->getCont(), '%', strlen(((SfxEntry*)sfx)->getCont()))
                    )                
               )
	     )
#endif
         ) {

	    // calculate syllable number of the word
	    numsyllable += get_syllable(st, i);

	    // + 1 word, if syllable number of the prefix > 1 (hungarian convention)
	    if (pfx && (get_syllable(((PfxEntry *)pfx)->getKey(),strlen(((PfxEntry *)pfx)->getKey())) > 1)) wordnum++;

	    // NEXT WORD(S)
	    rv_first = rv;
	    rv = lookup((word+i)); // perhaps without prefix

        // search homonym with compound flag
        while ((rv) && (rv->next_homonym) && 
			!((TESTAFF(rv->astr, compound, rv->alen)) ||
			  (compound3 && (TESTAFF(rv->astr, compound3, rv->alen))))) {
            rv = rv->next_homonym;
        }

	    oldnumsyllable2 = numsyllable;
	    oldwordnum2 = wordnum;

	    // fonev_morfo
#ifndef EXT_CLASS
	    if ((rv) && (TESTAFF(rv->astr, 'I', rv->alen)) && !(TESTAFF(rv->astr, 'J', rv->alen))) {
		numsyllable--;
	    }
#endif	    
	    // increment word number, if the second root has a compoundword flag
	    if ((rv) && (compoundword) && 
		(TESTAFF(rv->astr, compoundword, rv->alen))) {
		    wordnum++;
	    }

	    // check forbiddenwords
	    if ((rv) && (rv->astr) && (TESTAFF(rv->astr,forbiddenword,rv->alen))
		    && (! TESTAFF(rv->astr, onlyroot, rv->alen))) {
                        st[i] = ch;
                        continue;
            }
                    
	    // second word is acceptable, as a root?
	    // hungarian conventions: compounding is acceptable,
	    // when compound forms consist of 2 words, or if more,
	    // then the syllable number of root words must be 6, or lesser.
	    if ((rv) && (
	              TESTAFF(rv->astr, compound, rv->alen) ||
	              ((compound3 !=  FLAG_NULL) &&
		             TESTAFF(rv->astr, compound3, rv->alen))
	            )
		&& (
		      ((cpdwordmax==0) || (wordnum+1<cpdwordmax)) || 
		      ((cpdmaxsyllable==0) || 
		          (numsyllable+get_syllable(rv->word,rv->wlen)<=cpdmaxsyllable))
		    )
		&& (
		     (rv != rv_first)
		   )
		)
		 {
		      // bad compound word
                      strcat(*result, presult);
                      if (rv->description) {
                        if (*(rv->description) == '[') strcat(*result, rv->word);
                        strcat(*result, rv->description);
                      } else {
                        strcat(*result, word+i); // kell mg tkeress XXX
                        strcat(*result, "[STEM]");
                      }
                      strcat(*result, "\n");
		              ok = 1;
	    }

	    numsyllable = oldnumsyllable2 ;
	    wordnum = oldwordnum2;

	    // perhaps second word has prefix or/and suffix
            sfx = NULL;
	    sfxflag = FLAG_NULL;
	    rv = affix_check((word+i),strlen(word+i), compound);

            if (!rv && compound3) {
                sfx = NULL;
                pfx = NULL;
                rv = affix_check((word+i),strlen(word+i), compound3);
            }

            // check non_compound flag in suffix and prefix
            if ((rv) && 
                ((pfx && ((PfxEntry*)pfx)->getCont() &&
                    TESTAFF(((PfxEntry*)pfx)->getCont(), compoundforbidflag, 
                        strlen(((PfxEntry*)pfx)->getCont()))) ||
                (sfx && ((SfxEntry*)sfx)->getCont() &&
                    TESTAFF(((SfxEntry*)sfx)->getCont(), compoundforbidflag, 
                        strlen(((SfxEntry*)sfx)->getCont()))))) {
                    rv = NULL;
            }

	    // check forbiddenwords
	    if ((rv) && (rv->astr) && (TESTAFF(rv->astr,forbiddenword,rv->alen))
	    	    && (! TESTAFF(rv->astr, onlyroot, rv->alen))) {
                        st[i] = ch;
                        continue;
                    }

	    // calculate syllable number of the word
	    numsyllable += get_syllable(word + i, strlen(word + i));

            // - affix syllable num.
            if (sfxappnd) {
	        numsyllable -= get_syllable(sfxappnd, strlen(sfxappnd));
            }

	    // + 1 word, if syllable number of the prefix > 1 (hungarian convention)
	    if (pfx && (get_syllable(((PfxEntry *)pfx)->getKey(),strlen(((PfxEntry *)pfx)->getKey())) > 1)) wordnum++;

	    // increment syllable num, if last word has a SYLLABLENUM flag
	    // and the suffix is beginning `s'
#ifndef EXT_CLASS
	    if (cpdsyllablenum) {
	      switch (sfxflag) {
		case 'c': { numsyllable+=2; break; }
		case 'J': { numsyllable += 1; break; }
                case 'I': { if (rv && TESTAFF(rv->astr, 'J', rv->alen)) numsyllable += 1; break; }
	      }
	    }
#endif
	    // increment word number, if the second word has a compoundword flag
	    if ((rv) && (compoundword) && 
		(TESTAFF(rv->astr, compoundword, rv->alen))) {
		    wordnum++;
	    }
	    // second word is acceptable, as a word with prefix or/and suffix?
	    // hungarian conventions: compounding is acceptable,
	    // when compound forms consist 2 word, otherwise
	    // the syllable number of root words is 6, or lesser.
	    if ((rv) && 
                    (
		      ((cpdwordmax==0) || (wordnum+1<cpdwordmax)) || 
		      ((cpdmaxsyllable==0) || 
//		          (numsyllable+get_syllable(rv->word,rv->wlen)<=cpdmaxsyllable))
		          (numsyllable <= cpdmaxsyllable))
		    )
		&& (
		   (rv != rv_first)
		   )) {
                      char * m = affix_check_morph((word+i),strlen(word+i), compound);
                      if (*m == '\0') m = affix_check_morph((word+i),strlen(word+i), compound3);
                      strcat(*result, presult);
                      line_uniq(m);
                      if (strchr(m, '\n')) {
                            strcat(*result, "(");
                            strcat(*result, line_join(m, '|'));
                            strcat(*result, ")");
                      } else {
                            strcat(*result, m);
                      }
                      free(m);
                      strcat(*result, "\n");
		              ok = 1;
	    }

	    numsyllable = oldnumsyllable2;
	    wordnum = oldwordnum2;

	    // perhaps second word is a compound word (recursive call)
//	    if ((wordnum < maxwordnum)) {
	    if ((wordnum < maxwordnum) && (ok == 0)) {
			compound_check_morph((word+i),strlen(word+i), 
				wordnum+1, numsyllable, maxwordnum, 0, result, presult);
	    } else {
		rv=NULL;
	    }
	}
        st[i] = ch;
	wordnum = oldwordnum;
	numsyllable = oldnumsyllable;
    }
    return 0;
}    
#endif



// check word for suffixes

struct hentry * AffixMgr::suffix_check (const char * word, int len, 
       int sfxopts, AffEntry * ppfx, char ** wlst, int maxSug, int * ns, 
       const FLAG cclass, const FLAG needflag, int prefix_in_cpms)
{
    struct hentry * rv = NULL;

    // first handle the special case of 0 length suffixes
    SfxEntry * se = (SfxEntry *) sStart[0];

    while (se) {
       
       if (!cclass || se->getCont()) {
       rv = se->check(word,len, sfxopts, ppfx, wlst, maxSug, ns, (FLAG) cclass, needflag);
       if ((rv) && (prefix_in_cpms || se->getCont())) {
           return rv;
       }
       }
       se = se->getNext();
    }
  
    // now handle the general case
    unsigned char sp = *((const unsigned char *)(word + len - 1));
    SfxEntry * sptr = (SfxEntry *) sStart[sp];

    while (sptr) {
        if (isRevSubset(sptr->getKey(), word + len - 1, len)
        ) {
            if (((prefix_in_cpms && !cclass) || sptr->getCont()) &&
            (rv = sptr->check(word,len, sfxopts, ppfx, wlst, maxSug, ns, cclass, needflag))) {
                    sfx=(AffEntry *)sptr; // BUG: sfx not stateless
            	    sfxflag = sptr->getFlag(); // BUG: sfxflag not stateless
		    if (!sptr->getCont()) sfxappnd=sptr->getKey(); // BUG: sfxappnd not stateless
            	    return rv;
	    }
             sptr = sptr->getNextEQ();
        } else {
	     sptr = sptr->getNextNE();
        }
    }

    return NULL;
}

// check word for two-level suffixes

struct hentry * AffixMgr::suffix_check_twosfx(const char * word, int len, 
       int sfxopts, AffEntry * ppfx, const FLAG needflag)
{
    struct hentry * rv = NULL;

    // first handle the special case of 0 length suffixes
    SfxEntry * se = (SfxEntry *) sStart[0];
    while (se) {
#ifndef EXT_CLASS
        if (contclasses[(unsigned int)(se->getFlag())])
#endif
        {
            rv = se->check_twosfx(word,len, sfxopts, ppfx, needflag);
            if (rv) return rv;
        }
        se = se->getNext();
    }
  
    // now handle the general case
    unsigned char sp = *((const unsigned char *)(word + len - 1));
    SfxEntry * sptr = (SfxEntry *) sStart[sp];

    while (sptr) {
        if (isRevSubset(sptr->getKey(), word + len - 1, len)) {
#ifndef EXT_CLASS
            if (contclasses[(unsigned int)(sptr->getFlag())])
#endif
            {
	        rv = sptr->check_twosfx(word,len, sfxopts, ppfx, needflag); // (HU)
                if (rv) {
            	    sfxflag = sptr->getFlag(); // BUG: sfxflag not stateless
		    if (!sptr->getCont()) sfxappnd=sptr->getKey(); // BUG: sfxappnd not stateless
            	    return rv;
                }
            }
            sptr = sptr->getNextEQ();
        } else {
	     sptr = sptr->getNextNE();
        }
    }

    return NULL;
}

#ifdef HUNMORPH
char * AffixMgr::suffix_check_twosfx_morph(const char * word, int len, 
       int sfxopts, AffEntry * ppfx, const FLAG needflag)
{
    char result[MAXLNLEN];
    char result2[MAXLNLEN];
    char result3[MAXLNLEN];
    
    char * st;

    result[0] = '\0';
    result2[0] = '\0';
    result3[0] = '\0';

    // first handle the special case of 0 length suffixes
    SfxEntry * se = (SfxEntry *) sStart[0];
    while (se) {
#ifndef EXT_CLASS
        if (contclasses[(unsigned int)(se->getFlag())])
#endif
        {
            st = se->check_twosfx_morph(word,len, sfxopts, ppfx, needflag);
            if (st) {
                if (ppfx) {
                    if (((PfxEntry *) ppfx)->getMorph()) strcat(result, ((PfxEntry *) ppfx)->getMorph());
                    strcat(result,"+");
                }
                strcat(result, st);
                free(st);
                strcat(result, "+");
                if (se->getMorph()) strcat(result, se->getMorph());
                strcat(result, "\n");
            }
        }
        se = se->getNext();
    }
  
    // now handle the general case
    unsigned char sp = *((const unsigned char *)(word + len - 1));
    SfxEntry * sptr = (SfxEntry *) sStart[sp];

    while (sptr) {
        if (isRevSubset(sptr->getKey(), word + len - 1, len)) {
#ifndef EXT_CLASS
            if (contclasses[(unsigned int)(sptr->getFlag())]) 
#endif
            {
	        st = sptr->check_twosfx_morph(word,len, sfxopts, ppfx, needflag); // (HU)
                if (st) {
            	    sfxflag = sptr->getFlag(); // BUG: sfxflag not stateless
		    if (!sptr->getCont()) sfxappnd=sptr->getKey(); // BUG: sfxappnd not stateless
                    strcpy(result2, st);
                    free(st);

                    strcpy(result3, "+");
#ifdef DEBUG
#ifdef EXT_CLASS
                strcat(result3, sptr->getFlag());
#else
                strcat(result3, " ");
                result3[strlen(result3) - 1] = sptr->getFlag();
#endif
                strcat(result3, ":");
#endif
                if (sptr->getMorph()) strcat(result3, sptr->getMorph());
                strlinecat(result2, result3);
                strcat(result2, "\n");
                strcat(result,  result2);
                }
            }
            sptr = sptr->getNextEQ();
        } else {
	     sptr = sptr->getNextNE();
        }
    }
    if (result) return mystrdup(result);
    return NULL;
}


char * AffixMgr::suffix_check_morph(const char * word, int len, 
       int sfxopts, AffEntry * ppfx, const FLAG cclass, const FLAG needflag)
{
    char result[MAXLNLEN];
    
    struct hentry * rv = NULL;

    result[0] = '\0';

    // first handle the special case of 0 length suffixes
    SfxEntry * se = (SfxEntry *) sStart[0];
    while (se) {
       rv = se->check(word,len, sfxopts, ppfx, NULL, 0, 0, cclass, needflag);
       while (rv) {
           //return rv;
           if (ppfx) {
//                strcat(result,((PfxEntry *) ppfx)->getKey());
                if (((PfxEntry *) ppfx)->getMorph()) strcat(result, ((PfxEntry *) ppfx)->getMorph());
                strcat(result,"+");
            }
           if (rv->description && *(rv->description)=='[') strcat(result, rv->word);
           if (rv->description) strcat(result, rv->description);
           strcat(result, "+");
//           strcat(result, se->getAffix());
           if (se->getMorph()) strcat(result, se->getMorph());
           strcat(result, "\n");
           rv = se->get_next_homonym(rv, sfxopts, ppfx, cclass, needflag);
       }
       se = se->getNext();
    }
  
    // now handle the general case
    unsigned char sp = *((const unsigned char *)(word + len - 1));
    SfxEntry * sptr = (SfxEntry *) sStart[sp];

    while (sptr) {
        if (isRevSubset(sptr->getKey(), word + len - 1, len)) {
	    ////rv = sptr->check(word,len, sfxopts, ppfx);
	    rv = sptr->check(word,len, sfxopts, ppfx, NULL, 0, 0, cclass, needflag); // (HU)
            
            while (rv) {
            //sfxflag=sptr->getFlag();
		    //sfxappnd=sptr->getKey();
                    if (ppfx) {
//                        strcat(result,((PfxEntry * )ppfx)->getKey());
                        if (((PfxEntry *) ppfx)->getMorph()) strcat(result, ((PfxEntry *) ppfx)->getMorph());
                        strcat(result, "+");
                    }    
                    if (rv->description && *(rv->description)=='[') strcat(result, rv->word);
                    if (rv->description) strcat(result, rv->description);
                    strcat(result, "+");
//                    strcat(result, sptr->getAffix());
#ifdef DEBUG
#ifdef EXT_CLASS
                strcat(result, sptr->getFlag());
#else
                strcat(result, " ");
                result[strlen(result) - 1] = sptr->getFlag();
#endif
                strcat(result, ":");
#endif
                    if (sptr->getMorph()) strcat(result, sptr->getMorph());
                    strcat(result, "\n");
            	    //return rv;
                rv = sptr->get_next_homonym(rv, sfxopts, ppfx, cclass, needflag);
	    }
             sptr = sptr->getNextEQ();
        } else {
	     sptr = sptr->getNextNE();
        }
    }

    if (*result) return mystrdup(result);
    return NULL;
}
#endif

// check if word with affixes is correctly spelled
struct hentry * AffixMgr::affix_check (const char * word, int len, const FLAG needflag)
{
    struct hentry * rv= NULL;

    // check all prefixes (also crossed with suffixes if allowed) 
    rv = prefix_check(word, len, 1, needflag);
    if (rv) return rv;

    // if still not found check all suffixes
    rv = suffix_check(word, len, 0, NULL, NULL, 0, NULL, FLAG_NULL, needflag);

    if (havecontclass) {
        if (rv) return rv;

        // if still not found check all two-level suffixes
        rv = suffix_check_twosfx(word, len, 0, NULL, needflag);
        if (rv) return rv;
        // if still not found check all two-level suffixes
        rv = prefix_check_twosfx(word, len, 1, needflag);
    }

    return rv;
}

#ifdef HUNMORPH
// check if word with affixes is correctly spelled
char * AffixMgr::affix_check_morph(const char * word, int len, const FLAG needflag)
{
    char result[MAXLNLEN];
    char * st = NULL;

    *result = '\0';
    
    // check all prefixes (also crossed with suffixes if allowed) 
    st = prefix_check_morph(word, len, 1);
    if (st) {
        strcat(result, st);
        free(st);
    }

    // if still not found check all suffixes    
    st = suffix_check_morph(word, len, 0, NULL, '\0', needflag);
    if (st) {
        strcat(result, st);
        free(st);
    }

    if (havecontclass) {
        // if still not found check all two-level suffixes
        st = suffix_check_twosfx_morph(word, len, 0, NULL, needflag);
        if (st) {
            strcat(result, st);
            free(st);
        }

        // if still not found check all two-level suffixes
        st = prefix_check_twosfx_morph(word, len, 1, needflag);
        if (st) {
            strcat(result, st);
            free(st);
        }
    }
    
    return mystrdup(result);
}
#endif



// return length of replacing table
int AffixMgr::get_numrep()
{
  return numrep;
}

// return replacing table
struct replentry * AffixMgr::get_reptable()
{
  if (! reptable ) return NULL;
  return reptable;
}

// return replacing table
struct replentry * AffixMgr::get_accent()
{
  if (! accent ) return NULL;
  return accent;
}

// return text encoding of dictionary
char * AffixMgr::get_encoding()
{
  if (! encoding ) {
      encoding = mystrdup("ISO8859-1");
  }
  return mystrdup(encoding);
}

// return the preferred try string for suggestions
char * AffixMgr::get_try_string()
{
  if (! trystring ) return NULL;
  return mystrdup(trystring);
}

// return the preferred try string for suggestions
char * AffixMgr::get_wordchars()
{
  if (! wordchars ) return NULL;
  return mystrdup(wordchars);
}

// return the compound words control flag
FLAG AffixMgr::get_compound()
{
  return compound;
}

// return the forbidden words control flag
FLAG AffixMgr::get_forbiddenword()
{
  return forbiddenword;
}

// return the forbidden words flag modify flag
FLAG AffixMgr::get_onlyroot()
{
  return onlyroot;
}

// return the compound word signal flag
FLAG AffixMgr::get_compoundword()
{
  return compoundword;
}

// return the value of checknum
int AffixMgr::get_checknum()
{
  return checknum;
}

// return the value of prefix
const char * AffixMgr::get_prefix()
{
  return ((PfxEntry *)pfx)->getKey();
}

// return the value of suffix
const char * AffixMgr::get_suffix()
{
  return sfxappnd;
}

// return the value of suffix
const char * AffixMgr::get_version()
{
  return version;
}

// utility method to look up root words in hash table
struct hentry * AffixMgr::lookup(const char * word)
{
  if (! pHMgr) return NULL;
  return pHMgr->lookup(word);
}

// return the value of suffix
const int AffixMgr::have_contclass()
{
  return havecontclass;
}


/* parse in the try string */
int  AffixMgr::parse_try(char * line)
{
   if (trystring) {
      fprintf(stderr,"error: duplicate TRY strings\n");
      return 1;
   }
   char * tp = line;
   char * piece;
   int i = 0;
   int np = 0;
   while ((piece=mystrsep(&tp,' '))) {
      if (*piece != '\0') {
          switch(i) {
	      case 0: { np++; break; }
              case 1: { trystring = mystrdup(piece); np++; break; }
	      default: break;
          }
          i++;
      }
      free(piece);
   }
   if (np != 2) {
      fprintf(stderr,"error: missing TRY information\n");
      return 1;
   } 
   return 0;
}


/* parse in the name of the character set used by the .dict and .aff */
int  AffixMgr::parse_set(char * line)
{
   if (encoding) {
      fprintf(stderr,"error: duplicate SET strings\n");
      return 1;
   }
   char * tp = line;
   char * piece;
   int i = 0;
   int np = 0;
   while ((piece=mystrsep(&tp,' '))) {
      if (*piece != '\0') {
          switch(i) {
	     case 0: { np++; break; }
             case 1: { encoding = mystrdup(piece); np++; break; }
	     default: break;
          }
          i++;
      }
      free(piece);
   }
   if (np != 2) {
      fprintf(stderr,"error: missing SET information\n");
      return 1;
   } 
   return 0;
}


/* parse in the flag used by the controlled compound words */
int  AffixMgr::parse_cpdflag(char * line)
{
   if (compound !=  FLAG_NULL) {
      fprintf(stderr,"error: duplicate compound flags used\n");
      return 1;
   }
   char * tp = line;
   char * piece;
   int i = 0;
   int np = 0;
   while ((piece=mystrsep(&tp,' '))) {
      if (*piece != '\0') {
          switch(i) {
	     case 0: { np++; break; }
             case 1: { 
#ifdef EXT_CLASS
                    compound = mystrdup(piece);
#else
                    compound = *piece;
#endif
                    np++; break; 
                }

	     default: break;
          }
          i++;
      }
      free(piece);
   }
   if (np != 2) {
      fprintf(stderr,"error: missing compound flag information\n");
      return 1;
   }
   return 0;
}

/* parse in the first flag used by the controlled compound words */
int  AffixMgr::parse_cpdfirst(char * line)
{
   if (compound2 !=  FLAG_NULL) {
      fprintf(stderr,"error: duplicate compound first flags used\n");
      return 1;
   }
   char * tp = line;
   char * piece;
   int i = 0;
   int np = 0;
   while ((piece=mystrsep(&tp,' '))) {
      if (*piece != '\0') {
          switch(i) {
	     case 0: { np++; break; }
             case 1: {
#ifdef EXT_CLASS
                    compound2 = mystrdup(piece);
#else
                    compound2 = *piece;
#endif
                    np++; break;
                }
	     default: break;
          }
          i++;
      }
      free(piece);
   }
   if (np != 2) {
      fprintf(stderr,"error: missing compound first flag information\n");
      return 1;
   }
   return 0;
}

/* parse in the last flag used by the controlled compound words */
int  AffixMgr::parse_cpdlast(char * line)
{
   if (compound3 !=  FLAG_NULL) {
      fprintf(stderr,"error: duplicate compound last flags used\n");
      return 1;
   }
   char * tp = line;
   char * piece;
   int i = 0;
   int np = 0;
   while ((piece=mystrsep(&tp,' '))) {
      if (*piece != '\0') {
          switch(i) {
	     case 0: { np++; break; }
             case 1: {
#ifdef EXT_CLASS
                    compound3 = mystrdup(piece);
#else
                    compound3 = *piece;
#endif
                    np++; break;
                }
	     default: break;
          }
          i++;
      }
      free(piece);
   }
   if (np != 2) {
      fprintf(stderr,"error: missing compound last flag information\n");
      return 1;
   }
   return 0;
}

/* parse in the flag used by the compound_check() method */
int  AffixMgr::parse_cpdword(char * line)
{
   if (compoundword) {
      fprintf(stderr,"error: duplicate compoundword flags used\n");
      return 1;
   }
   char * tp = line;
   char * piece;
   int i = 0;
   int np = 0;
   while ((piece=mystrsep(&tp,' '))) {
      if (*piece != '\0') {
          switch(i) {
	     case 0: { np++; break; }
             case 1: { cpdwordmax = atoi(piece); np++; break; }
             case 2: { 
#ifdef EXT_CLASS
                    compoundword = mystrdup(piece);
#else
                    compoundword = *piece;
#endif
                    np++; break;
                }
	     default: break;
          }
          i++;
      }
      free(piece);
   }
   if (np < 2) {
      fprintf(stderr,"error: missing compound word information\n");
      return 1;
   }
   return 0;
}

/* parse in the flag used by the compound_check() method */
int  AffixMgr::parse_cpdforbid(char * line)
{
   if (compoundforbidflag) {
      fprintf(stderr,"error: duplicate compoundforbidflag flags used\n");
      return 1;
   }
   char * tp = line;
   char * piece;
   int i = 0;
   int np = 0;
   while ((piece=mystrsep(&tp,' '))) {
      if (*piece != '\0') {
          switch(i) {
	     case 0: { np++; break; }
             case 1: { 
#ifdef EXT_CLASS
                    compoundforbidflag = mystrdup(piece);
#else
                    compoundforbidflag = *piece;
#endif
                    np++; break;
                }

	     default: break;
          }
          i++;
      }
      free(piece);
   }
   if (np != 2) {
      fprintf(stderr,"error: missing compound forbidden flag information\n");
      return 1;
   }
   return 0;
}

/* parse in the flag used by forbidden words */
int  AffixMgr::parse_forbid(char * line)
{
   if (forbiddenword != '\0') {
      fprintf(stderr,"error: duplicate forbidden word flags used\n");
      return 1;
   }
   char * tp = line;
   char * piece;
   int i = 0;
   int np = 0;
   while ((piece=mystrsep(&tp,' '))) {
      if (*piece != '\0') {
          switch(i) {
	     case 0: { np++; break; }
             case 1: {
#ifdef EXT_CLASS
                    forbiddenword = mystrdup(piece);
#else
                    forbiddenword = *piece;
#endif
                    np++; break;
                }
	     default: break;
          }
          i++;
      }
      free(piece);
   }
   if (np < 2) {
      fprintf(stderr,"error: missing forbidden flag information\n");
      return 1;
   }
   return 0;
}

/* parse in the flag used modify FORBIDDENWORD */
int  AffixMgr::parse_onlyroot(char * line)
{
   if (onlyroot != '\0') {
      fprintf(stderr,"error: duplicate only root flags used\n");
      return 1;
   }
   char * tp = line;
   char * piece;
   int i = 0;
   int np = 0;
   while ((piece=mystrsep(&tp,' '))) {
      if (*piece != '\0') {
          switch(i) {
	     case 0: { np++; break; }
             case 1: {
#ifdef EXT_CLASS
                    onlyroot = mystrdup(piece);
#else
                    onlyroot = *piece;
#endif
                    np++; break;
                }             
	     default: break;
          }
          i++;
      }
      free(piece);
   }
   if (np < 2) {
      fprintf(stderr,"error: missing only root flag information\n");
      return 1;
   }
   return 0;
}



/* parse in the min compound word length */
int  AffixMgr::parse_cpdmin(char * line)
{
   char * tp = line;
   char * piece;
   int i = 0;
   int np = 0;
   while ((piece=mystrsep(&tp,' '))) {
      if (*piece != '\0') {
          switch(i) {
	     case 0: { np++; break; }
             case 1: { cpdmin = atoi(piece); np++; break; }
	     default: break;
          }
          i++;
      }
      free(piece);
   }
   if (np != 2) {
      fprintf(stderr,"error: missing compound min information\n");
      return 1;
   } 
   if ((cpdmin < 1) || (cpdmin > 50)) cpdmin = 3;
   return 0;
}

/* parse in the accent string for suggestaccent method */
int  AffixMgr::parse_accent(char * line)
{
   if (accent) {
      fprintf(stderr,"error: duplicate accent strings used\n");
      return 1;
   }
   char * tp = line;
   char * piece;
   int i = 0;
   int np = 0;
   while ((piece=mystrsep(&tp,' '))) {
      if (*piece != '\0') {
          switch(i) {
	     case 0: { np++; break; }
             case 1: {
	         accent = (replentry *) malloc(sizeof(struct replentry));
	         accent->pattern = mystrdup(piece); np++; break; }
             case 2: { accent->replacement = mystrdup(piece); np++; break; }
	     default: break;
          }
          i++;
      }
      free(piece);
   }
   if (np < 2) {
      fprintf(stderr,"error: missing compound word information\n");
      return 1;
   } else if (strlen(accent->pattern) != strlen(accent->replacement)) {
      fprintf(stderr,"error: the length of two accent strings not equal\n");
      return 1;      
   }
   return 0;
}

/* parse in the wordchars string */
int  AffixMgr::parse_wordchars(char * line)
{
   if (wordchars) {
      fprintf(stderr,"error: duplicate WORDCHARS strings\n");
      return 1;
   }
   char * tp = line;
   char * piece;
   int i = 0;
   int np = 0;
   while ((piece=mystrsep(&tp,' '))) {
      if (*piece != '\0') {
          switch(i) {
	      case 0: { np++; break; }
              case 1: { wordchars = mystrdup(piece); np++; break; }
	      default: break;
          }
          i++;
      }
      free(piece);
   }
   if (np != 2) {
      fprintf(stderr,"error: missing WORDCHARS information\n");
      return 1;
   } 
   return 0;
}


/* parse in the max syllablecount of compound words and  */
int  AffixMgr::parse_cpdsyllable(char * line)
{
   char * tp = line;
   char * piece;
   int i = 0;
   int np = 0;
   while ((piece=mystrsep(&tp,' '))) {
      if (*piece != '\0') {
          switch(i) {
	     case 0: { np++; break; }
             case 1: { cpdmaxsyllable = atoi(piece); np++; break; }
             case 2: { cpdvowels = mystrdup(piece);  np++; break; }
	     default: break;
          }
          i++;
      }
      free(piece);
   }
   if (np < 2) {
      fprintf(stderr,"error: missing compoundsyllable information\n");
      return 1;
   }
   if (np == 2) cpdvowels = "aeiouAEIOU";
   return 0;
}

/* parse in the flags, that increments syllable number */
int  AffixMgr::parse_syllablenum(char * line)
{
   char * tp = line;
   char * piece;
   int i = 0;
   int np = 0;
   while ((piece=mystrsep(&tp,' '))) {
      if (*piece != '\0') {
          switch(i) {
	     case 0: { np++; break; }
             case 1: { cpdsyllablenum = mystrdup(piece); np++; break; }
	     default: break;
          }
          i++;
      }
      free(piece);
   }
   if (np != 2) {
      fprintf(stderr,"error: missing cpdsyllablenum information\n");
      return 1;
   }
   return 0;
}

/* parse in the typical fault correcting table */
int  AffixMgr::parse_reptable(char * line, FILE * af)
{
   if (numrep != 0) {
      fprintf(stderr,"error: duplicate REP tables used\n");
      return 1;
   }
   char * tp = line;
   char * piece;
   int i = 0;
   int np = 0;
   while ((piece=mystrsep(&tp,' '))) {
       if (*piece != '\0') {
          switch(i) {
	     case 0: { np++; break; }
             case 1: { 
                       numrep = atoi(piece);
	               if (numrep < 1) {
			  fprintf(stderr,"incorrect number of entries in replacement table\n");
			  free(piece);
                          return 1;
                       }
                       reptable = (replentry *) malloc(numrep * sizeof(struct replentry));
                       np++;
                       break;
	             }
	     default: break;
          }
          i++;
       }
       free(piece);
   }
   if (np != 2) {
      fprintf(stderr,"error: missing replacement table information\n");
      return 1;
   } 
 
   /* now parse the numrep lines to read in the remainder of the table */
   char * nl = line;
   for (int j=0; j < numrep; j++) {
        fgets(nl,MAXLNLEN,af);
        mychomp(nl);
        tp = nl;
        i = 0;
        reptable[j].pattern = NULL;
        reptable[j].replacement = NULL;
        while ((piece=mystrsep(&tp,' '))) {
           if (*piece != '\0') {
               switch(i) {
                  case 0: {
		             if (strncmp(piece,"REP",3) != 0) {
		                 fprintf(stderr,"error: replacement table is corrupt\n");
                                 free(piece);
                                 return 1;
                             }
                             break;
		          }
                  case 1: { reptable[j].pattern = mystrdup(piece); break; }
                  case 2: { reptable[j].replacement = mystrdup(piece); break; }
		  default: break;
               }
               i++;
           }
           free(piece);
        }
	if ((!(reptable[j].pattern)) || (!(reptable[j].replacement))) {
	     fprintf(stderr,"error: replacement table is corrupt\n");
             return 1;
        }
   }
   return 0;
}

/* parse in the flag used by affix_check() */
int  AffixMgr::parse_hu_kotohangzo(char * line)
{
   if (hu_kotohangzo != '\0') {
      fprintf(stderr,"error: duplicate HU_KOTOHANGZO flags used\n");
      return 1;
   }
   char * tp = line;
   char * piece;
   int i = 0;
   int np = 0;
   while ((piece=mystrsep(&tp,' '))) {
      if (*piece != '\0') {
          switch(i) {
	     case 0: { np++; break; }
             case 1: { 
#ifdef EXT_CLASS
                    hu_kotohangzo = mystrdup(piece);
#else
                    hu_kotohangzo = *piece;
#endif
                    np++; break; 
                }
	     default: break;
          }
          i++;
      }
      free(piece);
   }
   if (np < 2) {
      fprintf(stderr,"error: missing HU_KOTOHANGZO flag information\n");
      return 1;
   }
   return 0;
}

/* parse in the version string */
int  AffixMgr::parse_version(char * line)
{
   if (version) {
      fprintf(stderr,"error: duplicate VERSION strings\n");
      return 1;
   }
   char * tp = line;
   char * piece = mystrsep(&tp,' ');
   version = mystrdup(tp);
   free(piece);
   return 0;
}

int  AffixMgr::parse_affix(char * line, const char at, FILE * af)
{
   int numents = 0;      // number of affentry structures to parse

#ifdef EXT_CLASS
   char achar[3] = "\0\0";
#else
   char achar='\0';      // affix char identifier
#endif

   short ff=0;
   struct affentry * ptr= NULL;
   struct affentry * nptr= NULL;

   char * tp = line;
   char * nl = line;
   char * piece;
   int i = 0;

   // split affix header line into pieces

   int np = 0;
   while ((piece=mystrsep(&tp,' '))) {
      if (*piece != '\0') {
          switch(i) {
             // piece 1 - is type of affix
             case 0: { np++; break; }
          
             // piece 2 - is affix char
             case 1: { 
                    np++; 
#ifdef EXT_CLASS
                    if (strlen(piece) > 1) {
                        achar[0] = piece[0];
                        achar[1] = piece[1];
                    } else {
                        achar[0] = '0';
                        achar[1] = piece[0];                    
                    }
#else
                    achar = *piece;
#endif
                    break; 
                    }
             // piece 3 - is cross product indicator 
             case 2: { np++; if (*piece == 'Y') ff = XPRODUCT; break; }

             // piece 4 - is number of affentries
             case 3: { 
                       np++;
                       numents = atoi(piece); 
                       ptr = (struct affentry *) malloc(numents * sizeof(struct affentry));
                       ptr->xpflg = ff;
#ifdef EXT_CLASS
                       strcpy(ptr->achar, achar);
#else
                       ptr->achar = achar;
#endif
                       break;
                     }

	     default: break;
          }
          i++;
      }
      free(piece);
   }
   // check to make sure we parsed enough pieces
   if (np != 4) {
#ifdef EXT_CLASS
       fprintf(stderr, "error: affix %s header has insufficient data in line %s\n",achar,nl);
#else
       fprintf(stderr, "error: affix %c header has insufficient data in line %s\n",achar,nl);
#endif
       free(ptr);
       return 1;
   }
 
   // store away ptr to first affentry
   nptr = ptr;

   // now parse numents affentries for this affix
   for (int j=0; j < numents; j++) {
      fgets(nl,MAXLNLEN,af);
      mychomp(nl);
      tp = nl;
      i = 0;
      np = 0;

      // split line into pieces
      while ((piece=mystrsep(&tp,' '))) {
         if (*piece != '\0') {
             switch(i) {

                // piece 1 - is type
                case 0: { 
                          np++;
                          if (nptr != ptr) nptr->xpflg = ptr->xpflg;
                          break;
                        }

                // piece 2 - is affix char
                case 1: { 
		          np++;
#ifdef EXT_CLASS
                          if (strcmp(piece, achar) != 0) {
                              fprintf(stderr, "error: affix %s is corrupt near line %s\n",achar,nl);
#else
                          if (*piece != achar) {
                              fprintf(stderr, "error: affix %c is corrupt near line %s\n",achar,nl);
#endif
                              fprintf(stderr, "error: possible incorrect count\n");
                              free(piece);
                              return 1;
                          }

#ifdef EXT_CLASS
                          if (nptr != ptr) strcpy(nptr->achar, ptr->achar);
#else
                          if (nptr != ptr) nptr->achar = ptr->achar;
#endif
                          break;
		        }

                // piece 3 - is string to strip or 0 for null 
                case 2: { 
                          np++;
                          nptr->strip = mystrdup(piece);
                          nptr->stripl = strlen(nptr->strip);
                          if (strcmp(nptr->strip,"0") == 0) {
                              free(nptr->strip);
                              nptr->strip=mystrdup("");
			      nptr->stripl = 0;
                          }   
                          break; 
                        }

                // piece 4 - is affix string or 0 for null
                case 3: { 
		          np++;
                          nptr->appnd = mystrdup(piece);
                          nptr->appndl = strlen(nptr->appnd);
                          if (strcmp(nptr->appnd,"0") == 0) {
                              free(nptr->appnd);
                              nptr->appnd=mystrdup("");
			      nptr->appndl = 0;
                          }   
                          break; 
                        }

                // piece 5 - is the conditions descriptions
                case 4: { 
                          np++;
                          encodeit(nptr,piece);

#ifdef HUNMORPH
                         nptr->morphcode = NULL;
#endif
                         nptr->contclass = NULL;
                         break;
                }
                
                // piece 6, 7 - continue classes, and morph. description
                case 5: {
		          np++;
#ifdef HUNMORPH
                          nptr->morphcode = mystrdup(piece);
#endif
                          break; 

                }

                case 6: {
                          np++;
                          nptr->contclass = mystrdup(piece);
                          havecontclass = 1;
#ifndef EXT_CLASS
                          for (char *p = piece; *p; p++) contclasses[(unsigned char)*p] = 1;
#endif
                          break;
                
                }

		default: break;
             }
             i++;
         }
         free(piece);
      }
      // check to make sure we parsed enough pieces
      if (np < 5) { // HUNMORPH
#ifdef EXT_CLASS
          fprintf(stderr, "error: affix %s is corrupt near line %s\n",achar,nl);
#else
          fprintf(stderr, "error: affix %c is corrupt near line %s\n",achar,nl);
#endif
          free(ptr);
          return 1;
      }
      nptr++;
   }
         
   // now create SfxEntry or PfxEntry objects and use links to
   // build an ordered (sorted by affix string) list
   nptr = ptr;
   for (int k = 0; k < numents; k++) {
      if (at == 'P') {
	  PfxEntry * pfxptr = new PfxEntry(this,nptr);
          build_pfxlist((AffEntry *)pfxptr);
      } else {
	  SfxEntry * sfxptr = new SfxEntry(this,nptr);
          build_sfxtree((AffEntry *)sfxptr); 
      }
      nptr++;
   }      
   free(ptr);
   return 0;
}




