#include "license.readme"

#include <sys/types.h>
#include <sys/stat.h>

#include <cstring>
#include <cstdlib>
#include <cstdio>

#include "langid.hxx"

// see langid.hxx for other options
#define MAXFILESIZE 1024*8      // size of file sample (in byte)
#define MAXWORD MAXFILESIZE/4   // max. word count
#define BINARY_LIMIT 0.01       // limit for binary characters (for recognize compressed data)
#define LANGWORD_LIMIT 0.1      // limit for output


//#define WINDOWS // Windows environment

#define HUNP_VERSION "0.1"

#ifdef WINDOWS
#define LIBDIR "C:\\MySpell"
#define HOME "C:\\MySpell"
#else
// Not Windows
#define LIBDIR "/usr/share/myspell"
#define HOME getenv("HOME")
#endif

extern char * mystrdup(const char * s);
extern int r_access(char * filename);

// array for words
char ** words;


int boot_word = BOOT_WORD; // language pre-recognition based on BOOT_WORD sampled words
int sample_word = SAMPLE_WORD; // sample size = SAMPLE_WORD words
int sample_lang = SAMPLE_LANG; // max. used languages after booting (0 = all languages)

void ident_file(LangId * pLI, char * filename)
{
    char buf[MAXFILESIZE + 1];
    int i;
    int j;
    int wnum;
    idententry * result;
    
    int size;
    
    struct stat st;
    if (! stat(filename, &st)) {
        size = (st.st_size < MAXFILESIZE) ? st.st_size : MAXFILESIZE;
    } else {
        fprintf(stderr,gettext("Can't open %s.\n"), filename);
        exit(1);
    }

    FILE *f;
    f = fopen(filename, "r");
    
    if (f) {
        // load file
        int binary_chars;
        fread(buf, size, 1, f);
        fclose(f);

        buf[size] = '\0';
        wnum = 0;

        // check non text characters in file        
        binary_chars = pLI->is_text(buf, size);     
           
        if (((float)binary_chars)/size > BINARY_LIMIT) {
            fprintf(stdout, gettext("BINARY"));
        } else {    

            // tokenize
            wnum = pLI->tokenize(words, buf, size, MAXWORD, LETTERS);
            
            if (wnum == 0) {
                printf(gettext("EMPTY"));
            } else {
                printf(gettext("words:%d; "), wnum);
                
                // detect language
                j = pLI->identify(&result, words, wnum, sample_word, boot_word, sample_lang);
        
                wnum = (BOOT_WORD + SAMPLE_WORD) > wnum ? wnum : (BOOT_WORD + SAMPLE_WORD);
                for(i = 0; (i < j); i++) {
                    if ((float)((result+i)->freq)/wnum >= LANGWORD_LIMIT) {
                        printf("%s:%.0f%%; ", (result+i)->lang, ((result+i)->freq)*100.0/wnum);
                    }
                    free((result+i)->lang);
                }                
                free(result);
                result = NULL;
            }
        }
        printf("\n");
        fflush(stdout);
    } else {
        fprintf(stderr,gettext(gettext("Can't open %s.\n")), filename);
        exit(1);
    }
}

void ident_filelist(LangId * pLI)
{
    char buf[MAXLNLEN];

    while(fgets(buf, MAXLNLEN, stdin)) {
        *(buf + strlen(buf) - 1) = '\0';
        fprintf(stdout, "%s: ", buf);
        ident_file(pLI, buf);
    }
}

int main(int argc, char** argv)
{
        LangId * pLI = NULL;

        int arg_files = -1; // first filename argumentum position in argv

        char * diclist_path = NULL;
        char * dic_path = NULL;
        char ** argstate = NULL;

#ifndef WINDOWS
        setlocale(LC_ALL, "");
        textdomain("hunp");
#endif

        for(int i=1; i<argc; i++) {
                if (argstate) {
                        *argstate = mystrdup(argv[i]);
                        argstate = NULL;
                } else if (strcmp(argv[i],"-d")==0) {
                    argstate = &diclist_path;
                    if (*argstate) free(*argstate);
                } else if (strcmp(argv[i],"-l")==0) {
                    argstate = &dic_path;
                    if (*argstate) free(*argstate);
                } else if ((strcmp(argv[i],"-h") == 0) || (strcmp(argv[i],"--help") == 0)) {
                        fprintf(stderr,"Hunp ");
                        fprintf(stderr,HUNP_VERSION);
                        fprintf(stderr,gettext(" - Language Identification Program\n"));
                        fprintf(stderr,gettext("Usage: hunp [-d diclist_file] [-l dictionary_path] [file(s)]\n"));
                        fprintf(stderr,gettext("Examples: hunp file.txt\n"));
                        fprintf(stderr,gettext("          hunp <filelist.txt\n"));
                        fprintf(stderr,gettext("          ls | hunp\n"));
                        exit(0);
                } else if ((strcmp(argv[i],"-v")==0) && (strcmp(argv[i],"--version")==0)) {
                        fprintf(stdout,HUNP_VERSION);
                        fprintf(stdout,"\n");
                        fprintf(stderr,gettext("Author: Nmeth Lszl, 2003, License: GNU GPL v2\n"));
                        fprintf(stderr,gettext("E-mail: szoszablya-user@lists.mokk.bme.hu\n"));
                        exit(0);
                } else if (strcmp(argv[i],"-a")==0) {
                        // check all words
                        sample_word = 1 << (sizeof(int) * 8 - 2);
                } else if ((arg_files==-1) && ((argv[i][0] != '-') && (argv[i][0] != '\0'))) {
                        arg_files = i;
                        if (! r_access(argv[i])) { // first check (before time-consuming dic. load)
                                fprintf(stderr,gettext("Can't open %s.\n"),argv[i]);
                                exit(1);
                        }
                }
        }

        words = (char **) malloc(MAXWORD * sizeof(char *));        
        if (! words) {
                fprintf(stderr, gettext("Insufficient memory."));
                exit(1);        
        }

        if (! diclist_path) {
                if (! (diclist_path=getenv("DICLIST"))) diclist_path = "dictionary.lst";
        }

        if (! dic_path) dic_path = LIBDIR;
        
        pLI = new LangId(diclist_path, dic_path);

        if (!pLI) {
                fprintf(stderr,gettext("Can't open dictionary list file.\n"));
                exit(1);
        }

        if (arg_files==-1) {
                ident_filelist(pLI);
        } else {
                for (int i = arg_files; i < argc; i++) {
                        fprintf(stdout, "%s: ", argv[i]);
                        ident_file(pLI, argv[i]);
                }
        }

        delete pLI;
        return 0;
}
