#include "license.readme"
#include "license_myspell.readme"

#include <cstring>
#include <cstdlib>
#include <cstdio>
#include <time.h>

#include "langid.hxx"

#ifndef WINDOWS
using namespace std;
#endif

int BINARY_CHARS[] = {
    1, //NUL '\0'
    1, //SOH
    1, //STX
    1, //ETX
    1, //EOT
    1, //ENQ
    1, //ACK
    0, //BEL '\a'
    0, //BS  '\b'
    0, //HT  '\t'
    0, //LF  '\n'
    0, //VT  '\v'
    0, //FF  '\f'
    0, //CR  '\r'
    1, //SO
    1, //SI
    1, //DLE
    1, //DC1
    1, //DC2
    1, //DC3
    1, //DC4
    1, //NAK
    1, //SYN
    1, //ETB
    1, //CAN
    1, //EM
    0, //SUB (MS-DOS ^Z)
    1, //ESC
    1, //FS
    1, //GS
    1, //RS
    1 //US
};

#define BINARY_CHARS_LEN (sizeof(BINARY_CHARS) / (sizeof(int)))

extern void mychomp(char * s);
extern char * mystrdup(const char *);

int compare(const void * i, const void * j);
int r_access(char * filename);

LangId::LangId(const char * diclist_path, const char * dic_path)
{
    /* random seed */
    srand((unsigned)time(NULL));

    /* set binary characters */
    for (int i = 0; i < 256; i++) {
        if ( ((unsigned int) i < BINARY_CHARS_LEN) && (BINARY_CHARS[(unsigned char)i])) {
            binary_chars[(unsigned char)i] = 1;
        } else {
            binary_chars[(unsigned char)i] = 0;
        }
    }

    /* set up the dict manager */
    pDMgr = new DictMgr(diclist_path, "DICT");
    
    if (!pDMgr) {
        char dl[PATHLEN];
        if (strlen(diclist_path) + strlen(dic_path) < PATHLEN) {
            sprintf(dl, "%s%s", dic_path, dic_path);
            pDMgr = new DictMgr(dl, "DICT");
        }
    }
    
    if (pDMgr) {
        dicn = pDMgr->get_list(&dicl);
        dics = (MySpell **) malloc(dicn * sizeof(MySpell*));

        /* load dictionaries */
        if (dics) {
            int i;
            char af[PATHLEN];
            char df[PATHLEN];
            for (i = 0; i < dicn; i++) {
                if (strlen((dicl+i)->filename) + 4 < PATHLEN) {
                    sprintf(af, "%s.aff", (dicl+i)->filename);
                    sprintf(df, "%s.dic", (dicl+i)->filename);
                }
                dics[i] = NULL;
                if (r_access(af) && r_access(df)) dics[i] = new MySpell(af, df);
                if (! dics[i]) {
                    if (strlen((dicl+i)->filename) + 4 + strlen(dic_path) < PATHLEN) {
                        sprintf(af, "%s%c%s.aff", dic_path, PATHSEP, (dicl+i)->filename);
                        sprintf(df, "%s%c%s.dic", dic_path, PATHSEP, (dicl+i)->filename);
                    }
                    dics[i] = new MySpell(af, df);
                }

                if (! dics[i]) {
                    fprintf(stderr,"Missing dictionary: %s\n", (dicl+i)->filename);
                }
            }
        } else {
                fprintf(stderr,"Error - Insufficient Memory\n");
        }
    }
}

LangId::~LangId()
{
    if (dics) {
        int i;
        for (i = 0; i < dicn; i++) {
            if (dics[i]) delete dics[i];
        }
        free(dics);
    }
    if (pDMgr) delete pDMgr;
}

// returns with number of binary characters in data
int LangId::is_text(char * data, int size) {
    char * end = data + size;
    int n = 0;
    while (data != end) {
        if (binary_chars[(unsigned char) *data]) n++;
        data++;
    }
    return n;
}    

// word tokenization in place
int LangId::tokenize(char ** words, char * data, int size, int maxword, char * wordchars) {
    char * end = data + size;
    int is_wordchars[256];
    int state = 0;
    int word = 0;
    for (int i = 0; i < 256; i++) is_wordchars[(unsigned char) i] = 0;
    while (*wordchars != '\0') {
        is_wordchars[(unsigned char) *wordchars] = 1;
        wordchars++;
    }
    while ((data != end) && (word < maxword)) {
        if (state) {
            if (! is_wordchars[(unsigned char) *data]) {
                state = 0;
                *data = '\0';
                
                // delete one letter word
                if ((data - words[word - 1] == 1)) word--;
            }
        } else {
            if (is_wordchars[(unsigned char) *data]) {
                state = 1;
                words[word++] = data;
            }
        }
        data++;
    }
    return word;
}    

// detect languages of the given word list with random sampling
int LangId::identify(idententry** llst, char ** wlst, int wnum,
            int sample_word, int boot_word, int sample_lang) {
    int i;
    int n = 0;
    char * randomword;
    idententry * ilst;
    
    *llst = (idententry*) malloc((dicn + 1) * sizeof(idententry));
    if (! *llst) {
        fprintf(stderr,gettext("Error - Insufficient Memory\n"));
        return -1;    
    }
    ilst = *llst;

    if (! ilst) return 0;
    
    int * dicq = (int *) malloc(dicn * sizeof(int));
    if (! dicq) {
        fprintf(stderr,gettext("Error - Insufficient Memory\n"));
        free(ilst);
        return -1;
    }

    for (i = 0; i < dicn; i++) {
        char * lang = (char *) malloc(strlen((dicl+i)->lang) + strlen((dicl+i)->region) + 2);
        if (lang) {
            sprintf(lang, "%s_%s", (dicl+i)->lang, (dicl+i)->region);
        }
        (ilst+i)->lang = lang;
        (ilst+i)->freq = 0;
        (ilst+i)->pMS = dics[i];
    }

    (ilst+i)->lang = mystrdup(gettext("UNKNOWN"));
    (ilst+i)->freq = 0;
    (ilst+i)->pMS = NULL;
    

    if (dics) {
    
        // boot sampling (use all dictionaries)
        while ((n < boot_word) && (n < sample_word) && (n < wnum)) {
            int unknown = 1;
            randomword = get_random(wlst, wnum, n);
            for (i = 0; i < dicn; i++) {
                if (((ilst+i)->pMS) && (((ilst+i)->pMS)->spell(randomword))) {
                    (ilst+i)->freq++;
                    unknown = 0;
#ifdef LOG
                    fprintf(stderr,"(%s:%s)",(ilst+i)->lang,randomword);
#endif
                }
            }
            if (unknown) (ilst + dicn)->freq++;
            n++;
        }

        qsort(ilst, dicn, sizeof(idententry), &compare);

        // sampling 2
        // stop word spelling after first match
        // use max. SAMPLE_LANG dictionaries
        while ((n < sample_word+ boot_word) && (n < wnum)) {
            int unknown = 1;
            randomword = get_random(wlst, wnum, n);
            for (i = 0; (i < dicn) && (i < sample_lang); i++) {
                if (((ilst+i)->pMS) && (((ilst+i)->pMS)->spell(randomword))) {
                    (ilst+i)->freq++;
                    unknown = 0;
#ifdef LOG
                    fprintf(stderr,"(%s:%s)",(ilst+i)->lang,randomword);
#endif
                    break;
                }
            }
            if (unknown) (ilst + dicn)->freq++;
            n++;
        }
        
        qsort(ilst, dicn + 1, sizeof(idententry), &compare);

    } else return 0;

    return dicn + 1;
}

// abbrevation for identify
int LangId::identify(idententry** llst, char ** wlst, int wnum) {
    return identify(llst, wlst, wnum, SAMPLE_WORD, BOOT_WORD, SAMPLE_LANG);
}

// simple tokenization and identify
int LangId::identify_data(char *** langlist, char * data, int size,
        int sample_word, int boot_word, int sample_lang) {
    int wnum = size / 3 + 1;
    idententry * result;
    char ** wlst = (char **) malloc(wnum * sizeof(char *));
    if (wlst) {
        int i;
        int j;
        wnum = tokenize(wlst, data, size, wnum, LETTERS);
        j = identify(&result, wlst, wnum, sample_word, boot_word, sample_lang);
        *langlist = (char **) malloc(j * sizeof(char *));
        i = 0;
        while ((i < j) && (i < SAMPLE_LANG) &&
            ((float)((result+i)->freq)/wnum >= LANG_LIMIT)) {
                (*langlist)[i] = (result+i)->lang;
                i++;
        }
        free(result);
        free(wlst);
        return i;
    } else {
        fprintf(stderr,gettext("Error - Insufficient Memory\n"));
        return -1;
    }
}

// simple tokenization and identify
int LangId::identify_data(char *** langlist, char * data, int size) {
        return identify_data(langlist, data, size, SAMPLE_WORD, BOOT_WORD, SAMPLE_LANG);
}

// selects a random word with a random permutation algorithm
char * LangId::get_random(char ** wlst, int wnum, int n) {
    char * s;
    int r = n + rand() % (wnum - n);
    s = wlst[r];
    wlst[r] = wlst[n];
    wlst[n] = s;
    return s;
}

// comparator function for sorting the results
int compare(const void * i, const void * j) {
    if (((idententry *) i)->freq < ((idententry *) j)->freq) {
        return 1;
    } else if (((idententry *) i)->freq > ((idententry *) j)->freq) {
        return -1;
    }
    return 0;
}

// access for read
int r_access(char * filename)
{
    FILE *f = fopen(filename,"r");
    if (!f) return 0;
    fclose(f);
    return 1;
}
